__all__ = ['MEANINGS', 'MEANINGS_ALL', 'MEANINGS_NON_CORE']

MEANINGS = {
    1: ['I', 'je', 'ego', 'yo'],
    2: ['you', 'tú', 'tu', 'thou', 'tu, vos'],
    3: ['we', 'nous', 'nosotros', 'we two', 'nos', 'we (incl.)'],
    11: ['one', 'uno', 'un', 'unus'],
    12: ['two', 'deux', 'duo', 'dos'],
    18: ['person', 'personne', 'persona', 'homo, quidam'],
    19: ['fish', 'poisson', 'pez', 'fisx', 'piscis'],
    21: ['dog', 'perro', 'canis', 'chien'],
    22: ['louse', 'piojo', 'pediculus', 'pou'],
    23: ['tree', 'arbor', 'arbre', 'árbol', 'arbol'],
    25: ['leaf', 'folium', 'hoja', 'feuille'],
    28: ['skin', 'peaux', 'cutis, pellis', 'piel', 'peau'],
    30: ['blood', 'sangre', 'sang', 'sanguis'],
    31: ['bone', 'os', '%bone', 'hueso'],
    34: ['horn', 'corne', 'cuerno'],
    39: ['ear', 'oreille', 'oreja', 'auris'],
    40: ['eye', 'ojo', 'oculus', 'oeil'],
    41: ['nose', 'nariz', 'nez', 'nasus, nares'],
    43: ['tooth', 'dent', 'diente', 'dens'],
    44: ['tongue', 'lingua', 'lengua', 'langue'],
    47: ['knee', 'rodilla', 'genou', 'genu'],
    48: ['hand', 'main', 'mano', 'manus'],
    51: ['breast', 'breasts', 'seno', 'sein', 'mamma'],
    53: ['liver', 'higado', 'liverp', 'hígado', 'el higado', 'foie'],
    54: ['drink', 'boire', 'drinK', 'beber', 'bibere, potare', 'dring'],
    57: ['see', 'voir', 'ver'],
    58: ['hear', 'oir', 'oír', 'audire, exaudire', 'entendre', 'to hear'],
    61: ['die', 'morir', 'mourir', 'mori, obire'],
    66: ['come', 'venire', 'venir'],
    72: ['sun', 'soleil', 'sol'],
    74: ['star', 'etoile', 'estrella', 'stella, sidus'],
    75: ['water', 'eau', 'agua', 'aqua'],
    77: ['stone', 'piedra', 'pierre', 'lapis, saxum'],
    82: ['fire', 'firek', 'ignis', 'feu', 'fuego'],
    85: ['path', 'camino', 'semitas, trames, via', 'route', 'road', 'sendero'],
    86: ['mountain', 'mons', 'montagne', 'cerro', 'montaña', 'moundain'],
    92: ['night', 'noche', 'nuit'],
    95: ['full', 'full(ness)', 'lleno/a', 'plenus', 'lleno', 'plein'],
    96: ['new', 'nouveaux', 'nuevo'],
    100: ['name', 'nombre', 'nom', 'name (noun)'],
}

MEANINGS_NON_CORE = {
    4: 'this',
    5: 'that',
    6: 'who',
    7: 'what',
    8: 'not',
    9: 'all',
    10: 'many',
    13: 'big',
    14: 'long',
    15: 'small',
    16: 'woman',
    17: 'man',
    20: 'bird',
    24: 'seed',
    26: 'root',
    27: 'bark',
    29: 'flesh',
    32: 'grease',
    33: 'egg',
    35: 'tail',
    36: 'feather',
    37: 'hair',
    38: 'head',
    42: 'mouth',
    45: 'claw',
    46: 'foot',
    49: 'belly',
    50: 'neck',
    52: 'heart',
    55: 'eat',
    56: 'bite',
    59: 'know',
    60: 'sleep',
    62: 'kill',
    63: 'swim',
    64: 'fly',
    65: 'walk',
    67: 'lie',
    68: 'sit',
    69: 'stand',
    70: 'give',
    71: 'say',
    73: 'moon',
    76: 'rain',
    78: 'sand',
    79: 'earth',
    80: 'cloud',
    81: 'smoke',
    83: 'ash',
    84: 'burn',
    87: 'red',
    88: 'green',
    89: 'yellow',
    90: 'white',
    91: 'black',
    93: 'hot',
    94: 'cold',
    97: 'good',
    98: 'round',
    99: 'dry',
}

MEANINGS_ALL = {}
MEANINGS_ALL.update({k: v[0] for k, v in MEANINGS.items()})
MEANINGS_ALL.update(MEANINGS_NON_CORE)
# Check consistency of the meanings:
assert set(MEANINGS_ALL.keys()) == set(range(1, 101))  # Keys are numbers from 1 to 100.
assert len(set(MEANINGS_ALL.values())) == 100  # Meaning labels are unique.



import itertools
import collections

from clldutils.apilib import API
from clldutils.misc import lazyproperty
from csvw.dsv import reader

from pyasjp import models


def iter_chunks(fp):
    wordlist = None

    for line in fp:
        m = models.LANGUAGE_LINE_PATTERN.match(line)
        if m:
            if wordlist:
                yield wordlist
            wordlist = [line]
        else:
            if wordlist:
                wordlist.append(line)

    if wordlist:
        yield wordlist


class ASJP(API):
    def to_txt(self, *doculects, **kw):
        res = [models.txt_header()]
        wals_family, wals_genus = None, None
        for dl in sorted(doculects, key=lambda d: (d.wals_family or 'ZZZ', d.wals_genus or 'ZZZ')):
            if dl.wals_family and dl.wals_family != wals_family:
                wals_marker = '3'
            elif dl.wals_genus and dl.wals_genus != wals_genus:
                wals_marker = '2'
            else:
                wals_marker = '1'
            wals_family = dl.wals_family
            wals_genus = dl.wals_genus
            res.append(dl.to_txt(
                add_missing=kw.get('add_missing', False),
                full_list=kw.get('full_list', False),
                wals_marker=wals_marker))
        return '\n'.join(res) + '\n'

    def iter_doculects(self, p=None):
        ids = set()
        p = p or self.repos / 'lists.txt'
        with p.open(encoding='cp1252') as fp:
            for chunk in iter_chunks(fp):
                d = models.Doculect.from_txt('\n'.join(chunk))
                if d.id in ids:  # pragma: no cover
                    print('skipping duplicate doculect {}'.format(d.id))
                    continue
                ids.add(d.id)
                yield d

    @lazyproperty
    def sources(self):
        res = collections.defaultdict(list)
        for i, row in enumerate(
            reader(self.repos / 'sources.csv', dicts=True, encoding='cp1252'),
            start=1,
        ):
            res[row['ASJP_NAME'].lower().replace('-', '_')].append(models.Source(
                id=i, **{k.lower(): v for k, v in row.items()}))
        return res

    def source(self, dl):
        return self.sources.get(dl.asjp_name.lower())

    def transcriber(self, dl):
        srcs = self.source(dl)
        if srcs:
            return [
                self.transcribers[trs] for trs in
                set(itertools.chain(*[src.list_made_by for src in srcs]))]

    @lazyproperty
    def transcribers(self):
        return {
            t.id: t for t in
            [models.Transcriber(*row) for row in reader(self.repos / 'transcribers.csv')]}



import re
import logging

import attr
from clldutils.misc import nfilter

from pyasjp.meanings import *  # noqa: F403

__all__ = [
    'Transcriber', 'Source', 'Word', 'Synset', 'Doculect', 'ASJPCODES', 'txt_header',
    'valid_strict_orthography',
]

# see
# Brown, Cecil H., Eric W. Holman, Søren Wichmann, and Viveka Vilupillai. 2008.
# Automated classification of the world’s languages:
# a description of the method and preliminary results.
# STUF – Language Typology and Universals 61:285-308.
CONSONANTS = 'pbfvmw8tdszcnrlSZCjT5ykgxNqXh7L4G!'
VOWELS = 'ieE3auo'
ASJPCODES = CONSONANTS + VOWELS
MODIFIERS = ''.join([
    "*",  # An asterisk following any one of the seven vowel symbols indicates vowel nasalization.
    "~",  # A modifier that follows two juxtaposed consonants.
    "$",  # Similar ~ except that it follows three juxtaposed consonants.
    '"',  # Immediately follows a consonant that is glottalized.
])
PUNCTUATION = " "
WORD_PATTERN = re.compile('[{}]+'.format(re.escape(ASJPCODES + MODIFIERS + PUNCTUATION)))
MISSING_WORD = 'XXX'
LANGUAGE_LINE_PATTERN = re.compile(
    r'(?P<name>[^{]+){(?P<w>[^|]*)\|(?P<e>[^@}]*)(@(?P<g>[^\}]*))?\}?')


def valid_strict_orthography(word):
    for modifier, modified in [
        ('*', VOWELS),
        ('~', CONSONANTS),
        ('$', CONSONANTS),
        ('"', CONSONANTS),
    ]:
        if re.search('(?<![{}]){}'.format(re.escape(modified), re.escape(modifier)), word):
            raise ValueError('Misplaced modifier: {}'.format(word))


@attr.s
class Transcriber:
    id = attr.ib()
    name = attr.ib()


@attr.s
class Source:
    id = attr.ib()
    asjp_name = attr.ib()
    author = attr.ib()
    year = attr.ib()
    title_etc = attr.ib()
    list_made_by = attr.ib(converter=lambda s: nfilter([ss.strip() for ss in re.split('/|->', s)]))


def valid_word(instance, attribute, value):
    if not WORD_PATTERN.fullmatch(value):
        raise ValueError('Invalid form: {}'.format(value))


@attr.s
class Word:
    form = attr.ib(validator=valid_word)
    loan = attr.ib(validator=attr.validators.instance_of(bool))

    @classmethod
    def from_txt(cls, txt):
        if txt.startswith('%'):
            return cls(form=txt[1:], loan=True)
        return cls(form=txt, loan=False)

    def __str__(self):
        return '%' + self.form if self.loan else self.form


@attr.s
class Synset:
    meaning_id = attr.ib(converter=int, validator=attr.validators.in_(MEANINGS_ALL.keys()))
    meaning = attr.ib()
    words = attr.ib()
    comment = attr.ib()

    @classmethod
    def from_txt(cls, line):
        header, body = line.split('\t', 1)
        body = body.strip()
        comment = ''
        if re.search('  | //', body):
            body, comment = re.split('  | //', body, 1)
        # Degenerate cases:
        elif re.fullmatch(r'[^/]+/\s*[^/]+/$', body):
            body, comment, _ = body.split('/')
            body = body.strip()
        elif body.endswith('//'):  # pragma: no cover
            body = body[:-2].strip()
        elif body.endswith('/'):  # pragma: no cover
            body = body[:-1].strip()
        comment = comment.strip()
        words = []
        for word in body.split(','):
            word = word.strip()
            if word and word != MISSING_WORD:
                try:
                    words.append(Word.from_txt(word))
                except ValueError:
                    logging.getLogger(__name__).warning('skipping invalid word "{}"'.format(word))

        number, meaning = header.split(maxsplit=1)
        if number.endswith('.'):
            number = number[:-1].strip()
        return cls(int(number), meaning.strip(), words, comment or None)

    @staticmethod
    def format(mid, meaning=None, form=MISSING_WORD, comment=None):
        return '{} {}\t{} //{}'.format(
            mid, meaning or MEANINGS_ALL[mid], form, ' ' + comment if comment else '')

    def __str__(self):
        return self.format(
            self.meaning_id,
            meaning=self.meaning,
            form=', '.join(str(w) for w in self.words),
            comment=self.comment)


def txt_header(synonyms=2, words=28, year=1700):
    """
     2    28  1700     3
(I4,20X,10A1)
    """
    def rjust(n, s=6):
        return str(n).rjust(s)

    lines = [
        '%s%s%s%s%s%s' % tuple(map(rjust, (synonyms, words, year, 3, 92, 72))), '(I4,20X,10A1)']

    for mid, concept in sorted(MEANINGS_ALL.items(), key=lambda p: p[0]):
        lines.append('%s%s%s' % (rjust(mid, 4), 20 * ' ', concept))

    lines.append('')
    for char in ASJPCODES:
        lines.append(char)
    lines.append('')
    lines.append('')
    return '\n'.join(lines)


@attr.s
class Doculect:
    id = attr.ib()
    name = attr.ib()
    classification_wals = attr.ib()
    classification_ethnologue = attr.ib()
    classification_glottolog = attr.ib()
    latitude = attr.ib()
    longitude = attr.ib()
    number_of_speakers = attr.ib()
    recently_extinct = attr.ib()
    long_extinct = attr.ib()
    year_of_extinction = attr.ib()
    code_wals = attr.ib()
    code_iso = attr.ib()
    synsets = attr.ib(converter=lambda v: [vv for vv in v if vv.words])

    @property
    def asjp_name(self):
        name = self.name
        if '"' in name:
            name = name.split('"')[1]
        return name.replace("'", "").replace(" ", "_").replace("/", "_").upper()

    @property
    def wals_family(self):
        return self.classification_wals.split('.')[0] if self.classification_wals else None

    @property
    def wals_genus(self):
        return self.classification_wals.split('.')[1] if self.classification_wals else None

    @classmethod
    def from_txt(cls, txt, **kw):
        """
        The second line gives properties of the languages, in fixed format separated by blanks
        (not tabs), so the columns are important.

        Col. 2: 3 if the language is the first one in a new WALS family, 2 if it’s the first
        language in a new WALS genus, 1 otherwise.

        Col. 4-10, right justified: latitude in degrees and
        hundredths of a degree; minus means South.

        Col. 12-18, right justified: longitude in degrees a
        nd hundredths of a degree; minus means West.
        Latitudes and longitudes were ascertained from WALS, or from the maps in Ethnologue or
        Moseley and Asher (1994), or from information in the source for the list.

        Col. 19-30, right justified: number of speakers, from Ethnologue. This number always
        refers to the entire language, as defined in Ethnologue, even if the list itself
        refers to a dialect. The number is

            0 if the number of speakers is unknown;
            -1 if the language is recently extinct;
            -2 if the language is long extinct;
            or if the approximate date of extinction is known, the date is preceded by a minus
            sign.

        In the ASJP software, if there is a date in the first line of the entire file, lists
        with earlier extinction dates here are ignored, as are lists with -2; otherwise, all
        lists are read.

        Col. 34-36: three-letter WALS code, if any.

        Col. 40-42: three-letter ISO639-3 code, if any. This code is included for languages in
        previous editions of Ethnologue even if they aren’t in the 17th edition. Languages
        that lack an ISO639-3 code but can be placed in the Ethnologue classification are
        given a code consisting of two letters followed by the number 0 (for use in ASJP
        software).
        """
        lines = nfilter(txt.split('\n'))
        m = LANGUAGE_LINE_PATTERN.match(lines[0])
        assert m
        line = lines[1]
        try:
            lat = line[4:11].strip()
            lng = line[11:18].strip()
            nos = int(line[18:30].strip() or 0)
            code_wals = line[33:36].strip() or None
            code_iso = line[39:42].strip() or None
        except ValueError:  # pragma: no cover
            _, lat, lng, nos, code_wals, code_iso = line.strip().split()
            nos = int(nos)
        return cls(
            id=m.group('name'),
            name=' '.join(s.capitalize() for s in m.group('name').split('_')),
            classification_wals=m.group('w') if m.group('w') else None,
            classification_ethnologue=m.group('e') if m.group('e') else None,
            classification_glottolog=m.group('g') if m.group('g') else None,
            latitude=float(lat) if lat else None,
            longitude=float(lng) if lng else None,
            number_of_speakers=nos if nos >= 0 else 0,
            recently_extinct=nos == -1 or nos < -2,
            long_extinct=nos == -2,
            year_of_extinction=abs(nos) if nos < -2 else None,
            code_wals=code_wals,
            code_iso=code_iso,
            synsets=[Synset.from_txt(line) for line in lines[2:] if '\t' in line],
        )

    def __str__(self):
        return self.to_txt()

    def to_txt(self, add_missing=False, full_list=False, wals_marker='1'):
        """render the wordlist in the ASJP plain text format.
        """
        nos = self.number_of_speakers
        if self.year_of_extinction:
            nos = -self.year_of_extinction
        elif self.recently_extinct:
            nos = -1
        elif self.long_extinct:
            nos = -2
        lines = [
            '%s{%s|%s@%s}' % (
                self.id,
                self.classification_wals or '',
                self.classification_ethnologue or '',
                self.classification_glottolog or ''),
            '%s%s%s%s%s%s' % (
                wals_marker.rjust(2),
                ('' if self.latitude is None else ('%.2f' % self.latitude)).rjust(8),
                ('' if self.longitude is None else ('%.2f' % self.longitude)).rjust(8),
                str(nos).rjust(12),
                (self.code_wals or '').rjust(6),
                (self.code_iso or '').rjust(6))]
        synsets = {s.meaning_id: s for s in self.synsets}
        for mid, concept in sorted(MEANINGS_ALL.items(), key=lambda i: i[0]):
            if mid in synsets:
                lines.append(str(synsets[mid]))
            elif add_missing and ((mid in MEANINGS) or full_list):
                lines.append(Synset.format(mid))
        return '\n'.join(lines)



"""
Diff synsets between two ASJP releases.
"""
import collections

from clldutils.clilib import PathType

from pyasjp import ASJP
from pyasjp.meanings import MEANINGS, MEANINGS_NON_CORE


def register(parser):
    parser.add_argument(
        'listA',
        type=PathType(type='file')
    )
    parser.add_argument(
        'listB',
        type=PathType(type='file')
    )
    parser.add_argument(
        '--match-label',
        help="Match synsets based on meaning label rather than meaning number",
        action='store_true',
        default=False,
    )


def run(args):
    api = ASJP('.')
    reverse_meaning_lookup = {}
    for mid, ms in MEANINGS.items():
        for m in ms:
            reverse_meaning_lookup[m] = mid
    for mid, m in MEANINGS_NON_CORE.items():
        reverse_meaning_lookup[m] = mid

    formsA = collections.defaultdict(lambda: collections.defaultdict(dict))
    for dl in api.iter_doculects(args.listA):
        for ss in dl.synsets:
            mid = reverse_meaning_lookup[ss.meaning] if args.match_label else ss.meaning_id
            formsA[dl.id][mid] = ', '.join(str(w) for w in ss.words)

    for dl in api.iter_doculects(args.listB):
        if dl.id in formsA:
            for ss in dl.synsets:
                mid = reverse_meaning_lookup[ss.meaning] if args.match_label else ss.meaning_id
                if mid in formsA[dl.id]:
                    formsB = ', '.join(str(w) for w in ss.words)
                    formsetB = set(
                        w.strip().replace('~', '') for w in formsB.split(','))
                    formsetA = set(
                        w.strip().replace('~', '') for w in formsA[dl.id][mid].split(','))
                    # if formsA[dl.id][mid] != formsB:
                    if formsetA != formsetB:
                        print('{}:{}:{} -> {}'.format(dl.id, mid, formsA[dl.id][mid], formsB))
